def dtype_detect(var) -> str:
    pass
    dtype=''
    return dtype


# def dtype_detect(var) -> Type:
#     if dtype == str:
#         dtype = ftype_detect(var)
#     else:
#         dtype=type(var)
#     return dtype