# from .dtype_detect import dtype_detect
# from .ftype_detect import ftype_detect
from .loader import FileLoader

from .utility import (
    all_equal,
    count_folder_number,
    count_file_number,
    create_nonduplicated_dict_key,
)
