

def lower_and_replace_space_with_underscore(input_string):
        return input_string.replace(
            '   ', '_'
        ).replace(
            '   ', '_'
        ).replace(
            '  ', '_'
        ).replace(
            ' ', '_'
        ).lower()

def upper_and_replace_space_with_underscore(input_string):
        return input_string.replace(
            '   ', '_'
        ).replace(
            '   ', '_'
        ).replace(
            '  ', '_'
        ).replace(
            ' ', '_'
        ).upper()